package com.streambase.contrib.list;

import java.util.*;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.DataType;
import com.streambase.sb.client.CustomFunctionResolver;


public class OneListOperations {

	/**
	 * Adds a list to a scalar.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b + A whose type is the same as List A
	 */
	@CustomFunctionResolver("ListCustomFunctionResolverBList")
	public static List<?> addL(Double b, List<?> A) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = b.doubleValue();
			result.add(B + (Double) o);
			while(itA.hasNext()){
				result.add(B + (Double) itA.next());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = b.intValue();
			result.add(B + (Integer) o);
			while(itA.hasNext()){
				result.add(B + (Integer) itA.next());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = b.longValue();
			result.add(B + (Long) o);
			while(itA.hasNext()){
				result.add(B + (Long) itA.next());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Adds a scalar to a list.
	 * 
	 * @param A A list of type integer, double or long
	 * @param b A scalar object of type integer, double or long
	 * @return A list representing A + b whose type is the same as List A
	 */
	@CustomFunctionResolver("ListCustomFunctionResolverAList")
	public static List<?> addR(List<?> A, Double b) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add((Double) o + B);
			while(itA.hasNext()){
				result.add((Double) itA.next() + B);
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add((Integer) o  + B);
			while(itA.hasNext()){
				result.add((Integer) itA.next() + B);
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add((Long) o  + B);
			while(itA.hasNext()){
				result.add((Long) itA.next() + B);
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Subtract a list from a from.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b - A whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverBList")
	public static List<?> subL(Double b, List<?> A) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add(B - (Double) o);
			while(itA.hasNext()){
				result.add(B - (Double) itA.next());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add(B - (Integer) o);
			while(itA.hasNext()){
				result.add(B - (Integer) itA.next());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add(B - (Long) o);
			while(itA.hasNext()){
				result.add(B - (Long) itA.next());
			}
			return result;
		}
		return new ArrayList<Object>();
	}

	/**
	 * Subtract a scalar from a list.
	 * 
	 * @param A A list of type integer, double or long
	 * @param b A scalar object of type integer, double or long
	 * @return A list representing A - b whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverAList")
	public static List<?> subR(List<?> A, Double b) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add((Double) o - B);
			while(itA.hasNext()){
				result.add((Double) itA.next() - B);
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add((Integer) o  - B);
			while(itA.hasNext()){
				result.add((Integer) itA.next() - B);
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add((Long) o  - B);
			while(itA.hasNext()){
				result.add((Long) itA.next() - B);
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Multiply a scalar by a list.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b*A whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverBList")
	public static List<?> mulL(Double b, List<?> A) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add(B * (Double) o);
			while(itA.hasNext()){
				result.add(B * (Double) itA.next());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add(B * (Integer) o);
			while(itA.hasNext()){
				result.add(B * (Integer) itA.next());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add(B * (Long) o);
			while(itA.hasNext()){
				result.add(B * (Long) itA.next());
			}
			return result;
		}
		return new ArrayList<Object>();
	}

	/**
	 * Multiply a list by a list.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing A*b whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverAList")
	public static List<?> mulR(List<?> A, Double b) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add((Double) o * B);
			while(itA.hasNext()){
				result.add((Double) itA.next() * B);
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add((Integer) o  * B);
			while(itA.hasNext()){
				result.add((Integer) itA.next() * B);
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add((Long) o  * B);
			while(itA.hasNext()){
				result.add((Long) itA.next() * B);
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Divide a scalar by a list.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b/A whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverBList")
	public static List<?> divL(Double b, List<?> A) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add(B / (Double) o);
			while(itA.hasNext()){
				result.add(B / (Double) itA.next());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add(B / (Integer) o);
			while(itA.hasNext()){
				result.add(B / (Integer) itA.next());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add(B / (Long) o);
			while(itA.hasNext()){
				result.add(B / (Long) itA.next());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Divide a list by a scalar.
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing A/b whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverAList")
	public static List<?> divR(List<?> A, Double b) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add((Double) o / B);
			while(itA.hasNext()){
				result.add((Double) itA.next() / B);
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer B = ((Number) b).intValue();
			result.add((Integer) o  / B);
			while(itA.hasNext()){
				result.add((Integer) itA.next() / B);
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add((Long) o  / B);
			while(itA.hasNext()){
				result.add((Long) itA.next() / B);
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * A scalar to the power of a list
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b**A whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverBList")
	public static List<?> powL(Double b, List<?> A) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add(Math.pow(B, (Double) o));
			while(itA.hasNext()){
				result.add(Math.pow(B, (Double) itA.next()));
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Double B = ((Number) b).doubleValue();
			result.add(((Number) Math.pow(B, (Double) o)).intValue());
			while(itA.hasNext()){
				result.add(((Number) Math.pow(B, (Integer) itA.next())).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add(((Number) Math.pow(B, (Double) o)).longValue());
			while(itA.hasNext()){
				result.add(((Number) Math.pow(B, (Integer) itA.next())).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}

	/**
	 * A list to the power of a scalar
	 * 
	 * @param b A scalar object of type integer, double or long
	 * @param A A list of type integer, double or long
	 * @return A list representing b**A whose type is the same as List A
	 */	
	@CustomFunctionResolver("ListCustomFunctionResolverAList")
	public static List<?> powR(List<?> A, Double b) {
		if (A.size() == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double B = ((Number) b).doubleValue();
			result.add(Math.pow((Double) o, B));
			while(itA.hasNext()){
				result.add(Math.pow((Double) itA.next(),B));
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Double B = ((Number) b).doubleValue();
			result.add(((Number) Math.pow((Double) o,B)).intValue());
			while(itA.hasNext()){
				result.add(((Number) Math.pow((Integer) itA.next(), B)).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long B = ((Number) b).longValue();
			result.add(((Number) Math.pow((Double) o,B)).longValue());
			while(itA.hasNext()){
				result.add(((Number) Math.pow((Integer) itA.next(), B)).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Return a list of successive differences.
	 * 
	 * @param A A list of type integer, double or long
	 * @return  A list representing differences between adjacent elements of A
	 * 	        whose type is the same as A.
	 */	
	@CustomFunctionResolver("DiffCustomFunctionResolver")
	public static List<?> diff(List<?> A) {
		if (A.size() == 0 || (A.size()-1) == 0){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			Double current, previous = (Double) o;
			while(itA.hasNext()){	
				current = (Double) itA.next(); 
				result.add(current - previous);
				previous = current;
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			Integer current, previous = (Integer) o;
			while(itA.hasNext()){
				current = (Integer) itA.next(); 
				result.add(current - previous);
				previous = current;
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			Long current, previous = (Long) o;
			while(itA.hasNext()){
				current = (Long) itA.next(); 
				result.add(current - previous);
				previous = current;
			}
			return result;
		}
		return new ArrayList<Object>();
	}	
	
	/**
	 * The custom function resolver for all the operations with two 
	 * parameters so that the above list functions type check at runtime.
	 */
	public static CompleteDataType ListCustomFunctionResolverAList(
			CompleteDataType A, CompleteDataType b) {
		CompleteDataType argA = A.getElementType();
		DataType argB = b.getDataType();		
		if (argA == null || argB != DataType.DOUBLE){return null;}
		if(argA == CompleteDataType.forInt())
			return CompleteDataType.forList(CompleteDataType.forInt());
		else if (argA == CompleteDataType.forDouble())
			return CompleteDataType.forList(CompleteDataType.forDouble());
		else if (argA == CompleteDataType.forLong())
			return CompleteDataType.forList(CompleteDataType.forLong());
		return null;
	}

	/**
	 * The custom function resolver for all the operations with two 
	 * parameters so that the above list functions type check at runtime.
	 */
	public static CompleteDataType ListCustomFunctionResolverBList(
			CompleteDataType b, CompleteDataType A) {
		CompleteDataType argA = A.getElementType();
		DataType argB = b.getDataType();
		if (argA == null || argB != DataType.DOUBLE){return null;}
		if(argA == CompleteDataType.forInt())
			return CompleteDataType.forList(CompleteDataType.forInt());
		else if (argA == CompleteDataType.forDouble())
			return CompleteDataType.forList(CompleteDataType.forDouble());
		else if (argA == CompleteDataType.forLong())
			return CompleteDataType.forList(CompleteDataType.forLong());
		return null;
	}

	/**
	 * The custom function resolver for all the difference operation so that the
	 * above list functions type check at runtime.
	 */
	public static CompleteDataType DiffCustomFunctionResolver(
			CompleteDataType A) {
		CompleteDataType argA = A.getElementType();
		if (argA == null){return null;}
		if(argA == CompleteDataType.forInt())
			return CompleteDataType.forList(CompleteDataType.forInt());
		else if (argA == CompleteDataType.forDouble())
			return CompleteDataType.forList(CompleteDataType.forDouble());
		else if (argA == CompleteDataType.forLong())
			return CompleteDataType.forList(CompleteDataType.forLong());
		return null;
	}
}
