package com.streambase.contrib.list;

import java.util.*;

import com.streambase.sb.CompleteDataType;
import com.streambase.sb.client.CustomFunctionResolver;

public class TwoListOperations {

	/**
	 * Adds two equally sized lists together.
	 * 
	 * @param A The first list of type integer, double, long or string
	 * @param B The second list of type integer, double, long or string
	 * @return A list representing A + B whose type is the same as List A
	 */
	@CustomFunctionResolver("VectorCustomFunctionResolverTwo")
	public static List<?> addLists(List<?> A, List<?> B) {
		if (A.size() != B.size()){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator(), itB = B.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			result.add((Double) o + ((Number) itB.next()).doubleValue());
			while(itA.hasNext()){
				result.add((Double) itA.next() + ((Number) itB.next()).doubleValue());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			result.add((Integer) o + ((Number) itB.next()).intValue());
			while(itA.hasNext()){
				result.add((Integer) itA.next() + ((Number) itB.next()).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			result.add((Long) o + ((Number) itB.next()).longValue());
			while(itA.hasNext()){
				result.add((Long) itA.next() + ((Number) itB.next()).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Subtracts two equally sized lists together.
	 * 
	 * @param A The first list of type integer, double or long
	 * @param B The second list of type integer, double or long
	 * @return A list representing A - B whose type is the same as List A
	 */
	@CustomFunctionResolver("VectorCustomFunctionResolverTwo")
	public static List<?> subLists(List<?> A, List<?> B) {
		if (A.size() != B.size()){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator(), itB = B.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			result.add((Double) o - ((Number) itB.next()).doubleValue());
			while(itA.hasNext()){
				result.add((Double) itA.next() - ((Number) itB.next()).doubleValue());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			result.add((Integer) o - ((Number) itB.next()).intValue());
			while(itA.hasNext()){
				result.add((Integer) itA.next() - ((Number) itB.next()).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			result.add((Long) o - ((Number) itB.next()).longValue());
			while(itA.hasNext()){
				result.add((Long) itA.next() - ((Number) itB.next()).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Multiplies two equally sized lists together.
	 * 
	 * @param A The first list of type integer, double or long
	 * @param B The second list of type integer, double or long
	 * @return A list representing A * B whose type is the same as List A
	 */
	@CustomFunctionResolver("VectorCustomFunctionResolverTwo")
	public static List<?> mulLists(List<?> A, List<?> B) {
		if (A.size() != B.size()){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator(), itB = B.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			result.add((Double) o * ((Number) itB.next()).doubleValue());
			while(itA.hasNext()){
				result.add((Double) itA.next() * ((Number) itB.next()).doubleValue());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			result.add((Integer) o * ((Number) itB.next()).intValue());
			while(itA.hasNext()){
				result.add((Integer) itA.next() * ((Number) itB.next()).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			result.add((Long) o * ((Number) itB.next()).longValue());
			while(itA.hasNext()){
				result.add((Long) itA.next() * ((Number) itB.next()).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}
	
	/**
	 * Divides two equally sized lists together. Note that one must 
	 * check manually that none of the elements of B are zero.
	 * 
	 * @param A The first list of type integer, double or long
	 * @param B The second list of type integer, double or long
	 * @return A list representing A / B whose type is the same as List A
	 */
	@CustomFunctionResolver("VectorCustomFunctionResolverTwo")
	public static List<?> divLists(List<?> A, List<?> B) {
		if (A.size() != B.size()){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator(), itB = B.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();
		if (o instanceof Double){
			ArrayList<Double> result = new ArrayList<Double>();
			result.add((Double) o / ((Number) itB.next()).doubleValue());
			while(itA.hasNext()){
				result.add((Double) itA.next() / ((Number) itB.next()).doubleValue());
			}
			return result;
		} else if (o instanceof Integer){
			ArrayList<Integer> result = new ArrayList<Integer>();
			result.add((Integer) o / ((Number) itB.next()).intValue());
			while(itA.hasNext()){
				result.add((Integer) itA.next() / ((Number) itB.next()).intValue());
			}
			return result;
		} else if (o instanceof Long){
			ArrayList<Long> result = new ArrayList<Long>();
			result.add((Long) o / ((Number) itB.next()).longValue());
			while(itA.hasNext()){
				result.add((Long) itA.next() / ((Number) itB.next()).longValue());
			}
			return result;
		}
		return new ArrayList<Object>();
	}

	/**
	 * Takes one list to the power of another. Note that one must 
	 * check manually that none of the elements of B are zero.
	 * 
	 * @param A The first list of type integer, double or long
	 * @param B The second list of type integer, double or long
	 * @return A list representing A^B whose type is the same as List A
	 */
	@CustomFunctionResolver("VectorCustomFunctionResolverTwo")
	public static List<?> powLists(List<?> A, List<?> B) {
		if (A.size() != B.size()){return new ArrayList<Object>();}
		Iterator<?> itA = A.iterator(), itB = B.iterator();
		if(!itA.hasNext()){return new ArrayList<Object>();}
		Object o = itA.next();	
		if (o instanceof Double || o instanceof Integer || o instanceof Long){
			ArrayList<Double> result = new ArrayList<Double>();
			result.add(Math.pow((Double) o, ((Number) itB.next()).doubleValue()));
			while(itA.hasNext()){
				result.add(Math.pow((Double) itA.next(),((Number) itB.next()).doubleValue()));
			}
			return result;
		}
		return new ArrayList<Object>();
	}

	/**
	 * The custom function resolver for all the operations so that the
	 * above list functions type check at runtime.
	 */
	public static CompleteDataType VectorCustomFunctionResolverTwo(
			CompleteDataType A, CompleteDataType B) {
		CompleteDataType argA = A.getElementType();
		CompleteDataType argB = B.getElementType();
		if (argA == null || argB == null){return null;}
		if (argB != CompleteDataType.forInt() && argB != CompleteDataType.forDouble() && argB != CompleteDataType.forLong()){return null;}
		if(argA == CompleteDataType.forInt())
			return CompleteDataType.forList(CompleteDataType.forInt());
		else if (argA == CompleteDataType.forDouble())
			return CompleteDataType.forList(CompleteDataType.forDouble());
		else if (argA == CompleteDataType.forLong())
			return CompleteDataType.forList(CompleteDataType.forLong());
		return null;
	}

}
